#!/usr/bin/perl -w

use Net::PcapUtils;
use NetPacket::Ethernet qw(:strip);
use NetPacket::IP qw(:strip);
use NetPacket::TCP;
use NetPacket::UDP;
use NetPacket::ICMP;

use English;
use Getopt::Long;
use Socket;
use IO::Handle;
use POSIX;

# Define version variables
my $vMajor = 0;
my $vMinor = 2;
my $vBuild = 19;

require "./conf/conf.inc";

# Define system variables
my $startIDS;
my %children;
my $children = 0;
my @sigs;
my $opt_c;
my $help;
my $opt_p;
my $opt_pp;
my $opt_q;
my $opt_qq;
my $opt_m;
my $opt_D;
my $opt_mm;
my $opt_pps;
my $opt_sig;
my $packetCount = 0;

# Parse Command Line Options
GetOptions (
		'start' => \$startIDS,
		'c=i' => \$opt_c,
		'help|?' => \$help,
		'pp' => \$opt_pp,
		'p' => \$opt_p,
		'mm' => \$opt_mm,
		'm' => \$opt_m,
		'pps' => \$opt_pps,
		'q' => \$opt_q,
		'qq' => \$opt_qq,
		'D' => \$opt_D,
		'sig' => \$opt_sig,
		);


launch();

################ FUNCTIONS ################

sub launch {

	my $run = 1;

	# Header
	print("\nAanval Intrusion Detection Engine (Spion)\nVersion: $vMajor.$vMinor.$vBuild\n---\n") unless ($opt_qq);

	# Exit immediately if we are not starting
	if (!$startIDS) { print "Nothing to do. Try -help\n"; $run = 0; }
	
	# Display Help
	if ($help) { displayHelp(); $run = 0; }

	if ($run == 1) {
		# Check for daemon mode
		if ($opt_D && $startIDS) { daemonRun(); }
		
		main();
	}

	# Footer
	print "---\nSpion - Done\n\n" unless ($opt_qq);
	
	exit;
}

sub main {

	@sigs = loadSignatures();

	my $countLimit = "";

	return 0, if (rootUserTest() == 0);
	
	if ($opt_c) {

	} 
	
    Net::PcapUtils::loop(\&process_pkt, SNAPLEN => 1024, PROMISC => 1, NUMPACKETS => -1);
    
}

sub printPacket {

	my ($sourceIP, $sourcePort, $destinationIP, $destinationPort) = @_;
	my $payload = "";

	if ($opt_q || $opt_qq) {
	
	} else {
		if ($opt_pp) {
			print "$sourceIP:$sourcePort -> $destinationIP:$destinationPort\n$payload";
		} else {
			print "$sourceIP:$sourcePort -> $destinationIP:$destinationPort\n";
		}
	}
}

sub rootUserTest {

	# Spion Must run as root
	if ($UID != 0 && $^O !~ /cygwin/i) {
	    print("ERROR ! Spion must run as root !\n");
	    return 0;
	} else {
		return 1;
	}
}

sub displayHelp {

	print "Usage: ./spion.pl [options]\n\n";
	print "-start       Start the IDS Engine\n";
	print "-p           Print Standard Packet Details\n";
	print "-pp          Print Standard Packet Details + Payload\n";
	print "-m           On Match, Print Standard Packet Details\n";
	print "-mm          On Match, Print Standard Packet Details + Payload\n";
	print "-c=x         Process x number of packets and exit\n";
	print "\n";
	print "-D           Daemon Mode - Runs in background\n";
	print "-q           Quiet Mode - Limited Output\n";
	print "-qq          Very Quiet Mode - No Output\n";
	print "-help        This Help Information\n";
	print "\n";

}

sub loadSignatures {

	my @filetmp;
	my @sigout;

	if (!@sigs) {
		print "Loading Signatures\n" unless ($opt_qq || $opt_q);
	} else {
		print "Signatures Refreshed\n" unless ($opt_qq || $opt_q);
	}
	
	if ((-e "$sig_base") && (-r "$sig_base")) {
	
		opendir(DIR, "$sig_base");
		my @branches = grep(/\.branch$/,readdir(DIR));
		closedir(DIR);

		foreach my $sigbranch (@branches) {

			opendir(DIR, "$sig_base$sigbranch");
			my @files = grep(/\.sig$/,readdir(DIR));
			closedir(DIR);
			
			foreach my $sigfile (@files) {
	
				my $status = 0;
				my $read = 0;
				my $sid = 0;
				my $bid = 0;
				my $content = 0;	
				my $name = 0;	
				my $desc = 0;
				my $proto = 0;
				my $case = 0;	
				my $sourcePort = 0;
				my $sourceIP = 0;
				my $destinationPort = 0;
				my $destinationIP = 0;
				my $flags = 0;
				my $syn = 0;
				my $fin = 0;
				my $psh = 0;
				my $ack = 0;
				my $rst = 0;
				my $urg = 0;
				my $rb1 = 0;
				my $rb2 = 0;
				my $ttl = 0;
				my $frag = 0;
				my $resp = 0;
				my $weight = 0;
				my $level = 0;
				my $rev = 0;
				my $cert = 0;
				my $auth = 0;
				my %signatures;
							
				if ((-e "$sig_base$sigbranch/$sigfile") && (-r "$sig_base$sigbranch/$sigfile")) {
					open(SIG, "$sig_base$sigbranch/$sigfile");
					@filetmp = <SIG>;
					
					# Break filename apart to get branch and signature id's
					my @tmparray = split(/\./, $sigfile);
					$bid = $tmparray[0];
					$sid = $tmparray[1];
					
					foreach my $line (@filetmp) {
						$line =~ s/\s+$//;

						if ($line =~ m/((?<=STATUS\=).+)/g) {
							$status = $&;
						} elsif ($line =~ m/((?<=READ\=).+)/g) {
							$read = $&;
						} elsif ($line =~ m/((?<=CONTENT\=).+)/g) {
							$content = $&;
						} elsif ($line =~ m/((?<=NAME\=).+)/g) {
							$name = $&;
						} elsif ($line =~ m/((?<=DESC\=).+)/g) {
							$desc = $&;
						} elsif ($line =~ m/((?<=PROTOCOL\=).+)/g) {
							$proto = $&;
						} elsif ($line =~ m/((?<=CASE\=).+)/g) {
							$case = $&;
						} elsif ($line =~ m/((?<=SRCPORT\=).+)/g) {
							$sourcePort = $&;
						} elsif ($line =~ m/((?<=DSTPORT\=).+)/g) {
							$destinationPort = $&;
						} elsif ($line =~ m/((?<=SRCIP\=).+)/g) {
							$sourceIP = $&;
						} elsif ($line =~ m/((?<=DSTIP\=).+)/g) {
							$destinationIP = $&;
						} elsif ($line =~ m/((?<=SYN\=).+)/g) {
							$syn = $&;
						} elsif ($line =~ m/((?<=FIN\=).+)/g) {
							$fin = $&;
						} elsif ($line =~ m/((?<=PSH\=).+)/g) {
							$psh = $&;
						} elsif ($line =~ m/((?<=ACK\=).+)/g) {
							$ack = $&;
						} elsif ($line =~ m/((?<=RST\=).+)/g) {
							$rst = $&;
						} elsif ($line =~ m/((?<=URG\=).+)/g) {
							$urg = $&;
						} elsif ($line =~ m/((?<=RB1\=).+)/g) {
							$rb1 = $&;
						} elsif ($line =~ m/((?<=RB2\=).+)/g) {
							$rb2 = $&;
						} elsif ($line =~ m/((?<=TTL\=).+)/g) {
							$ttl = $&;
						} elsif ($line =~ m/((?<=FRAG\=).+)/g) {
							$frag = $&;
						} elsif ($line =~ m/((?<=LEVEL\=).+)/g) {
							$level = $&;
						} elsif ($line =~ m/((?<=WEIGHT\=).+)/g) {
							$weight = $&;
						} elsif ($line =~ m/((?<=RESP\=).+)/g) {
							$resp = $&;
						} elsif ($line =~ m/((?<=REV\=).+)/g) {
							$rev = $&;
						} elsif ($line =~ m/((?<=CERT\=).+)/g) {
							$cert = $&;
						} elsif ($line =~ m/((?<=AUTH\=).+)/g) {
							$auth = $&;
						}
					}
					
					if ($status eq "ON") {
					
						push @sigout, "$bid.:.$sigbranch.:.$sid.:.$name.:.$content.:.$read.:.$case.:.$proto.:.$sourceIP.:.$sourcePort.:.$destinationIP.:.$destinationPort.:.$syn.:.$fin.:.$psh.:.$ack.:.$rst.:.$urg.:.$rb1.:.$rb2.:.$ttl.:.$frag.:.$level.:.$weight.:.$resp.:.$rev.:.$cert.:.$auth";
						
						print "\tLoaded [+]: $sigbranch->$bid.$sid ($name)\n" unless (@sigs || $opt_qq || $opt_q);
						
					} else {
						print "\tLoaded [-]: $sigbranch->$bid.$sid ($name)\n" unless (@sigs || $opt_qq || $opt_q);
					}
				}
			}
		}
	}
	
	print "\n---\n" unless (@sigs || $opt_qq || $opt_q);
	
	if ($opt_sig) {
		exit;
	} else {
		return @sigout;
	}

}

sub process_pkt {

    my($arg, $hdr, $pkt) = @_;

    my $ip_obj;
    my $p_obj;
	
	$packetCount++;
    
	$ip_obj = NetPacket::IP->decode(eth_strip($pkt));

	$ip_proto = $ip_obj->{'proto'};
	$ip_src_ip = $ip_obj->{'src_ip'};
	$ip_dest_ip = $ip_obj->{'dest_ip'};

	if ($ip_proto == 6) {
		$p_obj = NetPacket::TCP->decode(ip_strip(eth_strip($pkt)));
	} elsif ($ip_proto == 17) {
		$p_obj = NetPacket::UDP->decode(ip_strip(eth_strip($pkt)));
	} elsif ($ip_proto == 1) {
		$p_obj = NetPacket::ICMP->decode(ip_strip(eth_strip($pkt)));
	} else {
		return;
	}

	$p_src_port = $p_obj->{src_port};
	$p_dest_port = $p_obj->{dest_port};
	$p_payload = $p_obj->{data};
	
#	return unless ($ip_proto == 17 || $ip_proto == 6 || $ip_proto == 1);
	
	if ($opt_pp) {
		if ($ip_proto == 1) {
			print "Packet: $ip_proto $ip_src_ip -> $ip_dest_ip\n$p_payload\n";
		} else {
			print "Packet: $ip_proto $ip_src_ip:$p_src_port -> $ip_dest_ip:$p_dest_port\n$p_payload\n";
		}
	} elsif ($opt_p) {
		if ($ip_proto == 1) {
			print "Packet: $ip_proto $ip_src_ip -> $ip_dest_ip\n";
		} else {
			print "Packet: $ip_proto $ip_src_ip:$p_src_port -> $ip_dest_ip:$p_dest_port\n";
		}
	}	
	
	foreach my $string (@sigs) {

		my $negative = 0;
		my $positive = 0;
		my $trigger = 0;	
	
		my @tmparray = split(/\.:\./, $string);
		my $bid = $tmparray[0];
		my $bname = $tmparray[1];
		my $sid = $tmparray[2];
		my $sname = $tmparray[3];
		my $content = $tmparray[4];
		my $read = $tmparray[5];
		my $case = $tmparray[6];
		my $proto = $tmparray[7];
		my $srcIP = $tmparray[8];
		my $srcPort = $tmparray[9];
		my $dstIP = $tmparray[10];
		my $dstPort = $tmparray[11];
		my $syn = $tmparray[12];
		my $fin = $tmparray[13];
		my $psh = $tmparray[14];
		my $ack = $tmparray[15];
		my $rst = $tmparray[16];
		my $urg = $tmparray[17];
		my $rb1 = $tmparray[18];
		my $rb2 = $tmparray[19];
		my $ttl = $tmparray[20];
		my $frag = $tmparray[21];
		my $level = $tmparray[22];
		my $weight = $tmparray[23];
		my $resp = $tmparray[24];
		my $rev = $tmparray[25];
		my $cert = $tmparray[26];
		my $auth = $tmparray[27];
	
		if ($proto ne "OFF") {
			if ($proto && $negative == 0) {
				if ($ip_proto == $proto) {
					$positive = 1
				} else {
					$negative = 1;
				}
			}			
		}
			
		if ($dstPort && $negative == 0) {
			if ($p_dest_port) {
				if ($p_dest_port == $dstPort) {
					$positive = 1
				} else {
					$negative = 1;
				}
			} else {
				$negative = 1;
			}
		}			

		if ($srcPort && $negative == 0) {
			if ($p_src_port) {
				if ($p_src_port == $srcPort) {
					$positive = 1
				} else {
					$negative = 1;
				}
			} else {
				$negative = 1;
			}
		}		
		
		if ($content && $negative == 0) {
			$content =~ s/\s+$//;
			if ($p_payload =~ m/$content/g) {
				$positive = 1
			} else {
				$negative = 1;
			}
		}		
		
		# ALL
		# If negative == 0 then all matches were positive; trigger
		if ($negative == 0) {
			
			# Trigger				
			$trigger = 1;
		}
		
		if ($trigger == 1) {
		
			if ($ip_proto == 1) {

				if ($opt_mm) {
					print "Alert Match: [$bid.$sid] $ip_src_ip -> $ip_dest_ip\n$p_payload\n";
					
				} elsif ($opt_m) {
					print "Alert Match: [$bid.$sid] $ip_src_ip -> $ip_dest_ip\n";
				}
			
			} else {
		
				if ($opt_mm) {
					print "Alert Match: [$bid.$sid] $ip_src_ip:$p_src_port -> $ip_dest_ip:$p_dest_port\n$p_payload\n";
					
				} elsif ($opt_m) {
					print "Alert Match: [$bid.$sid] $ip_src_ip:$p_src_port -> $ip_dest_ip:$p_dest_port\n";
				}
			
			}
				
			writeMatch($ip_proto, $bid, $bname, $sid, $sname, $ip_src_ip, $p_src_port, $ip_dest_ip, $p_dest_port, $p_payload);
		}
	}			
	
	if ($packetCount >= 5000) {
		@sigs = loadSignatures();
		$packetCount = 0;
	}
	
}

sub writeMatch {
	
	# Data and payload
	my ($proto, $bid, $bname, $sid, $sname, $sourceIP, $sourcePort, $destinationIP, $destinationPort, $payload) = @_;
	
	# Random integers for filenames
	my $random = int( rand(10000)) + 20000;
	
	# If sensor .store does not exist -> create
	if (!-e "$alert_store$id" . ".store") {
		mkdir("$alert_store$id" . ".store");
	}
	
	# Create file name
	my $filename = time() . ".$random.$bid.$sid";
	
	# Epoch
	my $epoch = time();

	# Write DAT File
	open(FILEWRITE, "> $alert_store$id" . ".store/" . "$filename" . ".dat");
	if ($proto == 1) {
		print FILEWRITE "EPOCH:$epoch\nBID:$bid\nBNAME:$bname\nSID:$sid\nSNAME:$sname\nSIP:$sourceIP\nDIP:$destinationIP\n";
	} else {
		print FILEWRITE "EPOCH:$epoch\nBID:$bid\nBNAME:$bname\nSID:$sid\nSNAME:$sname\nSIP:$sourceIP\nSPORT:$sourcePort\nDIP:$destinationIP\nDPORT:$destinationPort\n";
	}
	close FILEWRITE;
	
	# Write Payload File (associated to DAT by base filename)
	open(FILEWRITE, "> $alert_store$id" . ".store/" . "$filename" . ".dat.pld");
	print FILEWRITE "$payload";
	close FILEWRITE;

}

sub ascii_to_hex ($) {

	(my $str = shift) =~ s/(.|\n)/sprintf("%02lx", ord $1)/eg;
	
	$str =~ s/\s+$//;
	
    return $str;
}

sub daemonRun {
	my $pid = fork; # Fork off the main process
	
	# If no PID is defined, the daemon failed to start
	defined ($pid) or die "Cannot start daemon: $!"; 
	
	# If we have a PID, the parent daemonized okay
	if ($pid) {
		print "Spion launched in daemon mode [$pid].\n" unless ($opt_qq);
	}
	
	exit if $pid; # Return control to the user
	
	POSIX::setsid(); # Become a session leader
	
	close (STDOUT); # Close file handles to detach from any terminals
	close (STDIN);
	close (STDERR);
	
	# Set up signals we want to catch. 
	
	# Any sort of death trigger results in instant death of all
	
	$SIG{HUP} = $SIG{INT} = $SIG{TERM} = sub {
		my $sig = shift;
		$SIG{$sig} = 'IGNORE';
		kill 'INT' => keys %children;
		die "killed by $sig\n";
		exit;
	};      
}
